/* umtsmon - a program to control/monitor your UMTS card
 * Copyright (C) 2006  Klaas van Gend
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation
 * applicable version is GPL version 2 only.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "umtsmon_global.h"

#include "mainwindow.h"
#include "TheDeviceManagerSingleton.h"
#include "SIMHandler.h"
#include "Runner.h"
#include "Popup.h"

#include "HelperList.h"

#include <qapplication.h>
#include <qmessagebox.h>
#include <assert.h>
#include <signal.h>
#include <stdlib.h>

#include <qtranslator.h>
#include <qtextcodec.h>

// forward declarations
void popupAbortMessageBox(int);
void printAbortMessage(int);


static const int MAXLONGCOMMANDSIZE  = 20;
static const int MAXSHORTCOMMANDSIZE =  4;

struct s_args
{
	char theFullCommand [MAXLONGCOMMANDSIZE];
	char theShortCommand[MAXSHORTCOMMANDSIZE];
	bool needsArgument;
	bool (*theFunctionPtr)(char* anArgument);
};


static bool setForceAutodetection( __attribute__ ((unused)) char* anArgument)
{
	TheDeviceManagerSingleton::me().forceAutodetection();
	return true;
}

static bool validateLocale(const QString& aLocale)
{
	// allowed formats are:   nl, nl_NL or nl_nl. 
	// No numbers, no dots, etc.
	QRegExp myRegExp_nl("^[a-z]{2}$");
	QRegExp myRegExp_nl_nl("^[a-z]{2}_[a-zA-Z0-9.\\-@]{0,15}$");
	if (myRegExp_nl.search(aLocale) == 0)
		return true;
	if (myRegExp_nl_nl.search(aLocale) == 0)
		return true;
	return false;
}


static QString theLanguage;
static bool setLang(char* anArgument)
{
	theLanguage = anArgument;
	
	if (validateLocale(theLanguage)==true)
		return true;

	printf("ERROR: you did not specify a valid locale.\n");	
	return false;
}


QTranslator* loadTranslator(const QString& aFileName, const QString& aLocale)
{
	DEBUG5("loadTranslator('%s', '%s')\n", aFileName.ascii(), aLocale.ascii());
	
	if (validateLocale(aLocale) == false && !aLocale.isEmpty())
	{
		DEBUG1("aLocale '%s' denied\n", aLocale.ascii());
		return NULL;
	}

	// Fix up the locale if required
	// i.e. pt_pt is not allowed, should be pt_PT :-(
	QString myLocale;
	if (aLocale.length()>3)
	{
		myLocale = aLocale.left(2).lower() + "_" + aLocale.right(aLocale.length()-3).upper();
	}
	else
		myLocale = aLocale.lower();
	
	QTranslator* myTranslator = new QTranslator(0);
	QString myQMFile = aFileName + "_" + myLocale;
	DEBUG5("Translation filename: '%s'\n", myQMFile.ascii());	
	// try in three locations: current dir, subdir i18n and a global dir.
	//  note that QTranslator itself already tries several permutations, 
	//  including aLocale equal to "de" if "de_AT" is not found  
	if (myTranslator->load( myQMFile, "."))
	{
		DEBUG5(" .. found in '.'\n");
		goto done;
	}
	if (myTranslator->load( myQMFile, "i18n"))
	{
		DEBUG5(" .. found in 'i18n'\n");
		goto done;
	}
	if (myTranslator->load( myQMFile, "/usr/share/" + aFileName + "/translations"))
	{
		DEBUG5(" .. found in system translations\n");
		goto done;
	}

	// failed:
	DEBUG5(" ... NOT FOUND\n");
	delete myTranslator;
	myTranslator = NULL;
	
	// just one extra: if "pt" doesn't work, try for "pt_PT"  
	if (aLocale.length() == 2)
		return loadTranslator(aFileName, aLocale + "_" + aLocale);
	
done:
	if (myTranslator != NULL)
		theLanguage = myLocale;
	return myTranslator;
}


static bool setserial(char* anArgument)
{
	DEBUG5("setserial: '%s'\n", anArgument);

	// no need to do advanced parameter checking
	//     - that'll be done later on, right? 
			
	// is there a comma in? in that case we neet to set both AT and PPP
	QStringList myList = QStringList::split(',', anArgument);
	
	switch (myList.count())
	{
		case 1:
			TheDeviceManagerSingleton::me().setSuggestedAT(myList[0]);
			TheDeviceManagerSingleton::me().setSuggestedPPP("");
			TheDeviceManagerSingleton::me().setSuggestedDeviceType(DeviceCapabilities::GENERIC_USER_SERIAL);
			break;
		case 2:
			TheDeviceManagerSingleton::me().setSuggestedAT(myList[0]);
			TheDeviceManagerSingleton::me().setSuggestedPPP(myList[1]);
			TheDeviceManagerSingleton::me().setSuggestedDeviceType(DeviceCapabilities::GENERIC_USER_SERIAL);
			break;
		default:
			printf("ERROR: supply max 2 devices to --serial\n");
			exit(2);
			break;
	}
	
	return true;
}

// oops global variable declared here...
int  theVerbosity = 1;
int  thePPPVerbosity = 0;

static bool setVerbosity(char* anArgument)
{
	DEBUG1("setVerbosity: '%s'\n", anArgument);
	
	// the argument should be a number. Let's parse.
	int myNumber = atoi(anArgument);
	if (myNumber < 1 || myNumber > MAX_VERBOSITY)
	{
		printf("ERROR: verbosity argument should be between 1 and 5\n");
		return false;
	}
	theVerbosity = myNumber;
	DEBUG2("set verbosity level to %d\n", theVerbosity);

	// check if we have a second argument after a comma (ppp-verbosity)
	char* comma = strchr(anArgument, ',');
	if (comma != NULL)
	{
		myNumber = atoi(comma + 1);
		if (myNumber < 0 || myNumber > 2)
		{
			printf("ERROR: PPP verbosity argument should be between 0 and 2\n");
			return false;
		}
		thePPPVerbosity = myNumber;
	}
	else
	{
		// verbosity > 2 => pppverbosity = 1; verbosity > 4 => pppverbosity = 2
		thePPPVerbosity = (theVerbosity - 1) / 2;
	}
	DEBUG2("set PPP verbosity level to %d\n", thePPPVerbosity);
	
	return true;
}


static bool help( __attribute__ ((unused)) char* anArgument)
{
	printf(APPNAME " " APPRELEASE "" APPFLAVOUR "\n\nhelp text\n\n");
	printf(" --help              gives this help text\n");
	printf(" -h                  gives this help text\n");
	printf(" --force-autodetect  force autodetection (don't use info from config file)\n");
	printf(" --serial <at>[,ppp] will prevent autodetection of devices and use serial port <at>[,<ppp>]\n");
	printf(" -s <at>[,<ppp>]     will prevent autodetection of devices and use serial port <at>[,<ppp>]\n");
	printf(" --verbosity <lvl>[,ppplvl] set verbosity, 1=little (default), %d=all\n", MAX_VERBOSITY);
	printf(" -v <lvl>[,ppplvl]   set verbosity\n");
	printf(" --language <locale> set the language, e.g. nl or pt_BR\n");
	printf("\n"); 
	exit(1);
}


static struct s_args theArgsTable[] =
{ 
	{ "--help", "-h", false, help, },
	{ "--serial", "-s", true, setserial, },
	{ "--verbosity", "-v", true, setVerbosity, }, 
	{ "--force-autodetect", "", false, setForceAutodetection, },
	{ "--language", "", true, setLang, },
//	keep this one last
	{ "\0", "\0", false, NULL, },
};

/** compare strings. 
 *  @returns 		If 0, no match, if positive nonzero: the character at this index is extra
 */
static int compareStrings(const char* anArgument, const char* aCommand)
{
	int i=0;
	while(true)
	{
		// we're at the end of the command - a match, maybe with info left...
		if (aCommand[i] == 0)
			return i;
		// the argument is shorter than the command - cannot be a match :-(
		if (anArgument[i] == 0)
			break;
		if (anArgument[i] != aCommand[i])
			break;
		i++;
	}

	// if we get here, it was no match :-(			
	return 0;
}


int main( int argc, char **argv )
{
	
	QApplication a( argc, argv );

	puts( APPNAME " version " APPRELEASE "" APPFLAVOUR " ." );

	// perform command line parsing
	// skip argv[0] because we know what's in there
	for ( int i = 1; i < a.argc(); i++ )
	{
		int  myMatchPoint = 0;
				
		char* myArgvPtr = a.argv()[i];
		struct s_args* myTEntry;

		for (int j = 0; theArgsTable[j].theFullCommand[0]!='\0'; j++)
		{
			myTEntry = &(theArgsTable[j]);

			myMatchPoint = compareStrings(myArgvPtr, myTEntry->theFullCommand);
			if (myMatchPoint)
				break;
			myMatchPoint = compareStrings(myArgvPtr, myTEntry->theShortCommand);
			if (myMatchPoint)
				break;
		} // end of the 'j' loop

		if (myMatchPoint)
		{
			char* anArgument = NULL;

			// there are several cases:
			//  1)  -s 3
			//  2)  -s3
			//  3)  -s=3
			//  4)  --serial=3
			//  5)  --serial 3
			//  6)  --help     (no arguments)
			//  7)  -h         (no arguments)

			// cases 1,2,3,4,5:
			if (myTEntry->needsArgument)
			{
				// is there extra data in this argument (i.e. cases 2,3,4)
				if (myArgvPtr[myMatchPoint] != 0)
				{
					// is it a '=' (case3,4) - then advance one character
					if (myArgvPtr[myMatchPoint] == '=')
						myMatchPoint++;
					anArgument = &myArgvPtr[myMatchPoint];
				}
				else
				{
					// case 1,5:

					// is there a next argument to parse?
					if ( (i+1) < a.argc())
					{
						anArgument = a.argv()[i+1];
						i++;
					}
					else
						anArgument = NULL;
				}
			}
			
			
			if (myTEntry->needsArgument==true && anArgument==NULL)
			{
				printf("parsing error: missing argument to '%s'!\n", myArgvPtr);
				exit(3);
			}
			if (myTEntry->theFunctionPtr(anArgument) == false)
			{
				printf("...\nExiting gracefully\n\n");
				help(NULL);
				// this should never be reached:
				exit(1);
			}
		}
		else
		{
			printf("ERROR: unknown argument '%s'\n", myArgvPtr);
			exit(2);
		}
	} // end of the 'i' loop

	// If a user who doesn't run umtsmon from the command line hits an assert,
	// the program stops without a visible UI message.
	// let's fix that here: if no verbosity is set, register a SIGABRT handler to pop up a dialog
	if (theVerbosity <= 1)
	{
		printf("installing GUI SIGABRT handler\n");
		signal(SIGABRT, popupAbortMessageBox);
	}
	else
	{
		// also register a handler explaining what to do 
		// if the users enabled lots of verbosity 
		if (theVerbosity >= 5)
		{
			printf("installing text SIGABRT handler\n");
			signal(SIGABRT, printAbortMessage);
		}		
	}


	// translation file for Qt, if it doesn't exist, we don't really care
	QTranslator* myQTTranslatorPtr = loadTranslator("qt", theLanguage);
	if (myQTTranslatorPtr == NULL)
		myQTTranslatorPtr = loadTranslator("qt", QTextCodec::locale());
	if (myQTTranslatorPtr != NULL)
		a.installTranslator( myQTTranslatorPtr );
	
	// translation file for application strings, if specified on command line we do care
	QTranslator* myAppTranslatorPtr = NULL;
	if (!theLanguage.isEmpty())
	{
		myAppTranslatorPtr = loadTranslator(APPNAME, theLanguage);
		if (myAppTranslatorPtr == NULL)
		{
			printf("You specified locale '%s' which wasn't found.\n", theLanguage.ascii());
			exit(1);
		}
	}
	if (myAppTranslatorPtr == NULL)
		myAppTranslatorPtr = loadTranslator(APPNAME, QTextCodec::locale());
	if (myAppTranslatorPtr != NULL)
	{
		DEBUG2("Installing Translator files for locale '%s'\n", theLanguage.ascii());
		a.installTranslator( myAppTranslatorPtr );
	}
	
	// If the user has specified command line parameters, these are already set
	// inside the Singleton
	TheDeviceManagerSingleton::getDeviceCapsRef();

	// at this point, we know where to expect the serial port(s)
	// let's do some checks on read/write access now - for the ports
	// and also suid checks on the helper applications we have
	HelperList myList1;
	if (myList1.testAndRun() == HelperList::QUIT)
		exit(0);
	
	// check for rootness - useful for the logs if users do odd things
	if (Runner::amIRoot())
		DEBUG1(APPNAME " is running as root!!!\n");

	{ // scope the SIMHandler because it no longer needs to exist after the askForPIN
	// Everthink fine arount the sim? SIMHandler will check it...
	SIMHandler mySIMHandler;
	mySIMHandler.askForPIN(); 
	}

	////////////////////////////////////
	// and start the application
	// do not forget to pass on some initialised classes
	mainwindow w;
	a.setMainWidget( &w );
	w.show();
	return a.exec();
}





void popupAbortMessageBox(__attribute__ ((unused)) int aSignalNumber)
{
	static bool isFirstTimeHere=true;

	if ( !isFirstTimeHere )
		return;
	isFirstTimeHere=false;

	QMessageBox::critical (NULL, 
			APPNAME " assert error",
			APPNAME " ran into a problem and unfortunately had to close. "
				"It is highly likely that if you run " APPNAME "from a command line, like:\n"
				APPNAME " -v5\n"
				"you will get information on what went wrong, including"
				" information on how to help us solve it.\n"
				"We thank you in advance for helping us improve " APPNAME "!",
			QMessageBox::Ok |  QMessageBox::Default, 
			QMessageBox::NoButton, 
			QMessageBox::NoButton);
	// do not call exit here - or the signal will not be handled correctly
}

void printAbortMessage(__attribute__ ((unused)) int aSignalNumber)
{
	printf("\n\n  *** " APPNAME " version " APPRELEASE "" APPFLAVOUR " closed due to an"
	           " unrecoverable program error.\n"
	       "  *** Please e-mail all of the above output to"
	           " umtsmon-develop@lists.sourceforge.net,\n"
	       "  *** Or go to the " APPNAME " website and file a bug.\n\n\n"); 
	// do not call exit here - or the signal will not be handled correctly
}
