/* umtsmon - a program to control/monitor your UMTS card
 * Copyright (C) 2006,2007,2008  Klaas van Gend
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation
 * applicable version is GPL version 2 only.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "umtsmon_global.h"
#include "HelperList.h"
#include "FileStuff.h"
#include "Popup.h"
#include "TheDeviceManagerSingleton.h"

#include "helperlistgui.h"

#include <qfile.h>
#include <qpixmap.h>

#include <assert.h>


/////////////////////////////////////////////////////////////////////////////
bool HelperData::fixup(void)
{
	DEBUG3("HelperApp::fixup()\n");
	return false;
}

/////////////////////////////////////////////////////////////////////////////
HelperApp::HelperApp(int aNumber, Runner::BinariesToRun aHelper,
		const QString& aVersion, const QString& aRemark) 
		: HelperData(aNumber), theHelper(aHelper), isBinaryMissing(false)
{
	Runner myRunner;
	
	theFullPath = myRunner.getPath(aHelper);
	isSUID   = myRunner.isSUID(aHelper)?(HelperData::YES):(HelperData::NO);
	theRemarks  = aRemark;
	theVersion  = aVersion;

	// It's debatable whether this code belongs here or in Runner...
	// For now it's here because this is the location where it concerns us
	switch(aHelper)
	{
	case Runner::RUNNERTESTER:
		theName = "INTERNAL ERROR";
		isSUID = HelperData::NOTAPPLICABLE;
		break;	
	case Runner::SOMESU:
		theName = "*su";
		isSUID = HelperData::NOTAPPLICABLE;
		break;
	case Runner::PPPD:
		theName = "pppd"; 	
		break;
	case Runner::CHAT:
		theName = "chat";
		break;
	case Runner::PCCARDCTL:
		theName = "(pc)cardctl";
		break;
	case Runner::USBMODESWITCH:
		theName = "USB mode switcher";
		break;
	}
	
	// do we see problems?
	if (theFullPath.isEmpty() || (!FileStuff::doesFileExist(theFullPath)))
	{
		isBinaryMissing = true;
		// binary isn't there :-(
		// in case of SOMESU and PPPD, that's killing
		// in case of PCCARDCTL or USBMODESWITCH, it may not even be relevant
		theFullPath = "!!! missing binary";
		switch(aHelper)
		{
		case Runner::SOMESU:
		case Runner::PPPD:
			// problem
			needsFix = true;
			break;
		case Runner::PCCARDCTL:
			// not a problem, we still didn't implement 'eject'
			break;
		case Runner::USBMODESWITCH:
			// let's put it this way: only if we find a binary
			// but it's not suid, it's a problem
			// if there's no binary, I don't care :-D
			break;
		case Runner::RUNNERTESTER:
		case Runner::CHAT:
			// *definitely* not a problem
			break;
		}
		DEBUG3("Binary %s is missing and that's %s.\n", theName.ascii(), needsFix?"major":"minor");
	}
	else
	{
		if (isSUID == NO && aHelper!=Runner::PCCARDCTL)
		{
			DEBUG1("binary '%s' is not SUID -> needs fixing\n", theFullPath.ascii());
			needsFix = true;
		}
	} 
}

void HelperApp::add(QListView* aListViewPtr)
{
	DEBUG3("HelperApp::add\n");
	QListViewItem* myQLVI = new QListViewItem (aListViewPtr, QString::number(theNumber,10), theName);
	myQLVI->setOpen(true);
	
	QListViewItem* mySubQLVI= new QListViewItem (myQLVI, "", 
				theFullPath, theVersion, "", theRemarks);
	switch(isSUID)
	{
	case HelperData::YES:
		{
			QPixmap myPixmap( QPixmap::fromMimeSource("ok.png") );
			mySubQLVI->setPixmap(3, myPixmap);
			break;
		}
	case HelperData::NO:
		{
			QPixmap myPixmap( QPixmap::fromMimeSource("process-stop-16x16.png") );
			mySubQLVI->setPixmap(3, myPixmap);
			break;
		}
	case HelperData::NOTAPPLICABLE:
		mySubQLVI->setText(3, "n/a");
		break;
	}

	// in case the binary is missing, let's put up a sign
	if (isBinaryMissing)
	{
		QPixmap myPixmap( QPixmap::fromMimeSource("process-stop-16x16.png") );
		mySubQLVI->setPixmap(1, myPixmap);
	}
	
}

bool HelperApp::fixup(void)
{
	DEBUG3("HelperApp::fixup()\n");
	
	if (isBinaryMissing)
	{
		Popup::Warning(APPNAME " couldn't find binary " + theName + 
			" which is required for full operation.");
		return false;
	}
	
	QString myCommand = "chown root:root " + theFullPath + " && chmod u+s " + theFullPath;
	DEBUG2("The fixup is: '%s'\n", myCommand.ascii());	
	Runner myRunner;
   	myRunner.runCommand(Runner::SOMESU,  myCommand);
	needsFix = myRunner.isSUID(theHelper)==false;
	isSUID = (!needsFix)?(HelperData::YES):(HelperData::NO);

	// needsFix is true when there's something to be fixed
	// we return here if everything was fixed - that's the opposite
	return !needsFix; 
}



/////////////////////////////////////////////////////////////////////////////
HelperDev::HelperDev(int aNumber, const QString& anATPort, const QString& aPPPPort) :
		HelperData(aNumber), theATPort(anATPort), thePPPPort(aPPPPort)
{			
	needsFix=false;
	if (FileStuff::isFileReadWrite(anATPort)!=FileStuff::READWRITE)
	{
		DEBUG1("Device %s is not read/write -> needs fix\n", anATPort.ascii());
		needsFix = true;
	}
	if (FileStuff::isFileReadWrite(aPPPPort)!=FileStuff::READWRITE)
	{
		DEBUG1("Device %s is not read/write -> needs fix\n", aPPPPort.ascii());
		needsFix = true;
	}
}


QListViewItem* HelperDev::newLVI(QListViewItem* aQLVI, 
				const QString& aPortName)
{
	QString myRemarks;
	bool isOK = false;
	
	if (FileStuff::doesFileExist(aPortName))
	{
		switch(FileStuff::isFileReadWrite(aPortName))
		{
		case FileStuff::READ:
			myRemarks = "user does not have write access";
			isOK = false;
			break;
		case FileStuff::WRITE:
			myRemarks = "user does not have read access";
			isOK = false;
			break;
		case FileStuff::READWRITE:
			myRemarks = "read/write access is OK";
			isOK = true;
			break;
		case FileStuff::NOTHING:
			myRemarks = "user does not have read nor write access "
			            "(or something else has a lock)";
			isOK = false;
			break;
		}
	}
	else
	{
		isOK = false;
		myRemarks = "not found on file system";
	}
		
	
	
	QListViewItem* mySubQLVI  = new QListViewItem (aQLVI, "", 
				aPortName, "n/a", "n/a", myRemarks);
	if (isOK)
	{
		QPixmap myPixmap( QPixmap::fromMimeSource("ok.png") );
		mySubQLVI->setPixmap(4, myPixmap);
	}
	else
	{
		QPixmap myPixmap( QPixmap::fromMimeSource("process-stop-16x16.png") );
		mySubQLVI->setPixmap(4, myPixmap);
	}
	return mySubQLVI;
}

void HelperDev::add(QListView* aListViewPtr)
{
	QListViewItem* myQLVI = new QListViewItem (aListViewPtr, 
				QString::number(theNumber, 10), 
				"Modem Communication Ports");
	myQLVI->setOpen(true);

	//FIXME, here are probably two small, one-time-only memory leaks:
	__attribute__ ((unused)) QListViewItem* mySubQLVI_AT  = newLVI(myQLVI, theATPort);
	__attribute__ ((unused)) QListViewItem* mySubQLVI_PPP = newLVI(myQLVI, thePPPPort);
}


bool HelperDev::fixDev(const QString& aPort)
{
	Runner myRunner;
	if (aPort.isEmpty())
		return false;
	if (FileStuff::isFileReadWrite(aPort)!=FileStuff::READWRITE)
	{
		QString myCommand = "chmod uog+rw " + aPort;
    	myRunner.runCommand(Runner::SOMESU,  myCommand);
	}
    return FileStuff::isFileReadWrite(aPort)==FileStuff::READWRITE;
}

bool HelperDev::fixup(void)
{
	DEBUG3("HelperDev::fixup()\n");
	bool theResult = true;

	if (fixDev(theATPort)==false)
		theResult = false;
	if (fixDev(thePPPPort)==false)
		theResult = false;

	needsFix = !theResult;
	return theResult;
}


// ###########################################################################
// ###########################################################################
// ###########################################################################
// ###########################################################################
// ###########################################################################



HelperList::TestResults HelperList::testAndRun(void)
{
	DEBUG3("HelperList::testAndRun(void)\n");

	createHelperList();
	
	// if nothing to fix - don't create window
	if (needsFixing()==false)
		return OK;

	// create gui and handle everything
	theGUIPtr = new helperlistgui();
	theGUIPtr->show();
	theGUIPtr->setHelperListPtr(this);
	addListToWidget();
	theGUIPtr->exec();
	return OK;
}

bool HelperList::needsFixing(void)
{
	DEBUG3("HelperList::needsFixing\n");
	HelperDataList::iterator i = theHelperDataList.begin(); 
	for (; i!=theHelperDataList.end(); i++)
	{
		if ( (*i)->isNeedingFix() )
			return true;
	}
	return false;
}


bool HelperList::fixIt(void)
{
	DEBUG3("HelperList::fixIt\n");
	
	bool wasAllFixed = true;
	
	HelperDataList::iterator i = theHelperDataList.begin(); 
	for (; i!=theHelperDataList.end(); i++)
	{
		if ( (*i)->isNeedingFix() )
		{
			if ( ((*i)->fixup()) == false)
				wasAllFixed = false;
		}
	}
	return wasAllFixed;
}

	
void HelperList::createHelperList(void)
{
	DEBUG3("HelperList::createHelperList()\n");
	// create the appropriate HelperAdata-derived structs
	theHelperDataList.push_back(new HelperApp(1, Runner::PPPD, "", "required if you want to make connections"));

	QString thePccardctlStdOut = "";
	if (!theRunner.getPath(Runner::PCCARDCTL).isEmpty())
	{
		theRunner.runCommand(Runner::PCCARDCTL, "--version");
		thePccardctlStdOut = theRunner.getStdOutStringList()[0];
	}
	theHelperDataList.push_back(new HelperApp(2, Runner::PCCARDCTL, 
					thePccardctlStdOut,
					"only required for PCMCIA cards"));
	
	theHelperDataList.push_back(new HelperApp(3, Runner::SOMESU));

 	theHelperDataList.push_back(new HelperApp(4, Runner::USBMODESWITCH,
					"","if this app is present, you probably need it."));

	theHelperDataList.push_back(new HelperDev(5, 
			TheDeviceManagerSingleton::getDeviceCapsRef().getATSerialPortName(), 
			TheDeviceManagerSingleton::getDeviceCapsRef().getPPPSerialPortName()));
	
	// CHAT is currently not used yet, but the code is below
//	theHelperDataList.push_back(new HelperApp(6, Runner::CHAT, "", 
//					"chat is not used for now"));
}

void HelperList::addListToWidget(void)
{
	theGUIPtr->clearList();
	HelperDataList::iterator i = theHelperDataList.begin(); 
	for (; i!=theHelperDataList.end();i++)
		theGUIPtr->add(*i);
}

